import logging
import sys

formatter = logging.Formatter('[%(levelname)s]: %(asctime)s - %(name)s - %(message)s')

koppeltaal_logger = logging.getLogger(__name__)
koppeltaal_logger.setLevel(logging.DEBUG)

console_logger = logging.StreamHandler(sys.stdout)
console_logger.setLevel(logging.DEBUG)
console_logger.setFormatter(formatter)

file_logger = logging.FileHandler(r'koppeltaal.log')  # FIXME: make optional
file_logger.setLevel(logging.INFO)
file_logger.setFormatter(formatter)

koppeltaal_logger.addHandler(console_logger)
koppeltaal_logger.addHandler(file_logger)


class NoopLogger:
    def critical(self, msg, **kwargs):
        ...

    def debug(self, msg, **kwargs):
        ...

    def error(self, msg, **kwargs):
        ...

    def exception(self, msg, **kwargs):
        ...

    def fatal(self, msg, **kwargs):
        ...

    def info(self, msg, **kwargs):
        ...

    def log(self, msg, **kwargs):
        ...

    def warning(self, msg, **kwargs):
        ...

    def warn(self, msg, **kwargs):
        ...


noop_koppeltaal_logger = NoopLogger()



class KoppeltaalApiError(Exception):
    """Base exception class for Koppeltaal errors."""

    ...


class OAuthError(Exception):
    """Base exception class for oauth errors."""

    ...


class GetAccessTokenError(OAuthError):
    pass


class AuthorizeError(OAuthError):
    pass



from contextlib import contextmanager
from enum import Enum
from functools import wraps
from logging import Logger
from typing import Any, Optional, Type, cast

import httpx
from fhir.resources.R4B.bundle import Bundle
from fhir.resources.R4B.operationoutcome import OperationOutcome
from pydantic import BaseModel


# HTTP methods helper
class Methods(Enum):
    GET = 'get'
    POST = 'post'
    PUT = 'put'
    PATCH = 'patch'
    DELETE = 'delete'


# Decorator for binding fhir urls to client methods and set expected response model
def url(
    url,
    response_model: Type[BaseModel],
    many: bool = False,
):
    def inner(func):
        @wraps(func)
        async def async_wrapper(self, *args, **kwargs):
            base_url = self.fhir_url
            kwargs['url'] = base_url + url

            # request token
            if not self.token or self.token.is_expired:
                await self.get_access_token()

            # make request
            response = await func(self, *args, **kwargs)

            # Response is empty, so we handling update, create or delete
            if not response:
                return None

            if many:
                # handle Bundle response; cast as Bundle than map to response model
                data = cast(Bundle, response)
                bundle = Bundle(**data)
                self.logger.info(data)

                if not bool(bundle.total):
                    return []

                return [
                    response_model(**cast(response_model, entry.resource).dict())
                    for entry in bundle.entry
                ]

            data = cast(response_model, response)
            return response_model(**data)

        return async_wrapper

    return inner


def handle_operation(response: str | bytes) -> OperationOutcome:
    try:
        return OperationOutcome(**response)
    except:
        return None


@contextmanager
def handle_response(response: httpx.Response, method: Methods, logger: Logger):
    try:
        print(response.content)
        response.raise_for_status()
    except (
        httpx.ConnectError,
        httpx.RemoteProtocolError,
        httpx.TransportError,
    ) as exception:
        logger.error(f"Connection error detected: {exception}")

    except httpx.RequestError as exception:
        logger.error(f'An error occurred while requesting {exception.request.url!r}')
    except httpx.HTTPStatusError as exception:
        logger.error(
            f'An error status_code returned {exception.request.url!r}: '
            f'{exception.response.status_code}'
        )

    logger.info(f"{method.name} - {response.request.url} - {response.status_code}")

    json_response = response.json()

    if operation := handle_operation(json_response):
        logger.error(operation)
        yield None
        return None

    yield json_response


async def make_request(
    client: httpx.AsyncClient,
    *,
    method: Methods,
    url: str,
    logger: Any,
    params: Optional[Any] = None,
    data: Optional[Any] = None,
):
    """Make Httpx request and handle errors"""

    if method == Methods.GET:
        response: httpx.Response = await client.get(url=url, params=params)
        with handle_response(response, method, logger) as response_data:
            return response_data
    elif method == Methods.DELETE:
        response = await client.delete(url=url, params=params)
        with handle_response(response, method, logger) as response_data:
            return response_data
    elif method == Methods.PUT:
        response = await client.put(url=url, params=params, data=data)
        with handle_response(response, method, logger) as response_data:
            return response_data
    elif method == Methods.POST:
        response = await client.post(url=url, params=params, data=data)
        with handle_response(response, method, logger) as response_data:
            return response_data

    return None



import asyncio
import json

import click
from pydantic import BaseModel
from rich import print

from kt2 import __version__ as VERSION
from kt2.client import KoppeltaalSDK
from kt2.helpers import font
from kt2.settings import CommandLineArgs, ConfigFile, load_settings, merge_config
from kt2.types import (
    KT2ActivityDefinitionList,
    KT2AuditEventList,
    KT2CareTeamList,
    KT2EndpointList,
    KT2OrganizationList,
    KT2PatientList,
    KT2PracticionerList,
    KT2SubscriptionList,
    KT2TaskList,
)

# ----------------------------------------------------------------------------
# KoppeltaalAsyncRunner: runs KoppeltaalSDK in aysnc-mode
# ----------------------------------------------------------------------------


def print_json(data):
    print(data)


class JsonResponse(BaseModel):
    __root__: KT2AuditEventList | KT2ActivityDefinitionList | KT2CareTeamList | KT2CareTeamList | KT2EndpointList | KT2OrganizationList | KT2PatientList | KT2PracticionerList | KT2TaskList | KT2SubscriptionList


class KoppeltaalAsyncRunner:
    def __init__(self, config: ConfigFile, debug: bool):
        self.config: ConfigFile = config

        # build SDK
        self.client = KoppeltaalSDK(
            client_name=self.config.client_name,
            fhir_url=self.config.fhir_url,
            oauth_token_url=self.config.oauth_token_url,
            oauth_authorize_url=self.config.oauth_authorize_url,
            smart_config_url=self.config.smart_config_url,
            domain=self.config.domain,
            client_id=self.config.client_id,
            oauth_introspection_token_url=self.config.oauth_introspection_token_url,
            jwks_keys=self.config.jwks_keys,
            enable_logger=True,
        )

    async def get_patients(self) -> None:
        """Get all Patient resources from Koppeltaal"""
        response = await self.client.get_patients()
        print_json(
            JsonResponse.parse_obj([obj for obj in response]).json(
                exclude_unset=True, indent=4
            )
        )

    async def get_patient(self, id: str) -> None:
        """Get single Patient resource from Koppeltaal"""
        response = await self.client.get_patient(id=id)
        print_json(
            JsonResponse.parse_obj([obj for obj in response]).json(
                exclude_unset=True, indent=4
            )
        )

    async def get_practitioners(self) -> None:
        """Get all Practitioner resources from Koppeltaal"""
        response = await self.client.get_practitioners()
        print_json(
            JsonResponse.parse_obj([obj for obj in response]).json(
                exclude_unset=True, indent=4
            )
        )

    async def get_practitioner(self, id: str) -> None:
        """Get single Practitioner resource from Koppeltaal"""
        response = await self.client.get_practitioner(id=id)
        print_json(
            JsonResponse.parse_obj([obj for obj in response]).json(
                exclude_unset=True, indent=4
            )
        )

    async def get_careteams(self) -> None:
        """Get all CareTeam resources from Koppeltaal"""
        response = await self.client.get_careteams()
        print_json(
            JsonResponse.parse_obj([obj for obj in response]).json(
                exclude_unset=True, indent=4
            )
        )

    async def get_careteam(self, id: str) -> None:
        """Get single CareTeam resource from Koppeltaal"""
        response = await self.client.get_careteam(id=id)
        print_json(
            JsonResponse.parse_obj([obj for obj in response]).json(
                exclude_unset=True, indent=4
            )
        )

    async def get_audit_events(self) -> None:
        """Get all AuditEvent resources from Koppeltaal"""
        response = await self.client.get_audit_events()
        print_json(
            JsonResponse.parse_obj([obj for obj in response]).json(
                exclude_unset=True, indent=4
            )
        )

    async def get_devices(self) -> None:
        """Get all Device resources from Koppeltaal"""
        response = await self.client.get_devices()
        print_json(
            JsonResponse.parse_obj([obj for obj in response]).json(
                exclude_unset=True, indent=4
            )
        )

    async def get_endpoints(self) -> None:
        """Get all Endpoint resources from Koppeltaal"""
        response = await self.client.get_endpoints()
        print_json(
            JsonResponse.parse_obj([obj for obj in response]).json(
                exclude_unset=True, indent=4
            )
        )

    async def get_endpoint(self, id: str) -> None:
        """Get single Endpoint resource from Koppeltaal"""
        response = await self.client.get_endpoint(id=id)
        print_json(
            JsonResponse.parse_obj([obj for obj in response]).json(
                exclude_unset=True, indent=4
            )
        )

    async def get_activitydefinitions(self) -> None:
        """Get all ActivityDefinition resources from Koppeltaal"""
        response = await self.client.get_activitydefinitions()
        print_json(
            JsonResponse.parse_obj([obj for obj in response]).json(
                exclude_unset=True, indent=4
            )
        )

    async def get_activitydefinition(self, id: str) -> None:
        """Get single ActivityDefinition resource from Koppeltaal"""
        response = await self.client.get_activitydefinition(id=id)
        print_json(
            JsonResponse.parse_obj([obj for obj in response]).json(
                exclude_unset=True, indent=4
            )
        )

    async def delete_activitydefinition(self, id: str) -> None:
        """Get single ActivityDefinition resource from Koppeltaal"""
        response = await self.client.delete_activitydefinition(id=id)
        print_json(
            JsonResponse.parse_obj([obj for obj in response]).json(
                exclude_unset=True, indent=4
            )
        )

    async def get_tasks(self) -> None:
        """Get all Task resources from Koppeltaal"""
        response = await self.client.get_tasks()
        print_json(
            JsonResponse.parse_obj([obj for obj in response]).json(
                exclude_unset=True, indent=4
            )
        )

    async def get_task(self, id: str) -> None:
        """Get single Task resource from Koppeltaal"""
        response = await self.client.get_task(id=id)
        print_json(
            JsonResponse.parse_obj([obj for obj in response]).json(
                exclude_unset=True, indent=4
            )
        )

    async def get_subscriptions(self) -> None:
        """Get all Subscription resources from Koppeltaal"""
        response = await self.client.get_subscriptions()
        print_json(
            JsonResponse.parse_obj([obj for obj in response]).json(exclude_none=True)
        )

    async def get_info(
        self,
    ) -> None:
        """Show information for Koppeltaal Api"""
        info = await self.client.get_info()
        print(info)


class KoppeltaalCli:
    """Koppeltaal Api Cli wrapper"""

    def __init__(self, **kwargs) -> None:
        # Get commandline arguments
        self.args = CommandLineArgs(**kwargs)
        # Get toml file settings
        self.toml_config = load_settings(self.args.config)
        # build config
        self.config = merge_config(self.toml_config, self.args.dict())
        # build runner
        self.runner = KoppeltaalAsyncRunner(self.config, self.args.debug)

    # ----------------------------------------------------------------------------#
    # FHIR Endpoints                                                              #
    # ----------------------------------------------------------------------------#

    def patients(self) -> None:
        asyncio.run(self.runner.get_patients())

    def patient(self, id: str) -> None:
        asyncio.run(self.runner.get_patient(id=id))

    def practitioners(self) -> None:
        asyncio.run(self.runner.get_practitioners())

    def practitioner(self, id: str) -> None:
        asyncio.run(self.runner.get_practitioner(id=id))

    def audit_events(self) -> None:
        asyncio.run(self.runner.get_audit_events())

    def devices(self) -> None:
        asyncio.run(self.runner.get_devices())

    def endpoints(self) -> None:
        asyncio.run(self.runner.get_endpoints())

    def endpoint(self, id: str) -> None:
        asyncio.run(self.runner.get_endpoint(id=id))

    def activitydefinitions(self) -> None:
        asyncio.run(self.runner.get_activitydefinitions())

    def activitydefinition(self, id: str) -> None:
        asyncio.run(self.runner.get_activitydefinition(id=id))

    def delete_activitydefinition(self, id: str) -> None:
        asyncio.run(self.runner.delete_activitydefinition(id=id))

    def tasks(self) -> None:
        asyncio.run(self.runner.get_tasks())

    def task(self, id: str) -> None:
        asyncio.run(self.runner.get_task(id=id))

    def subscriptions(self) -> None:
        asyncio.run(self.runner.get_subscriptions())

    # ----------------------------------------------------------------------------#
    # Function for info, version, config, and other usefull insights              #
    # ----------------------------------------------------------------------------#

    def info(self) -> None:
        asyncio.run(self.runner.get_info())

    def show_version(self) -> None:
        click.secho("Koppeltaal CLI tools")
        click.secho(font.renderText("Koppeltaal"), fg="blue")
        click.secho(f"Version {VERSION}")


pass_koppeltaal = click.make_pass_decorator(KoppeltaalCli, ensure=True)



import click
from pyfiglet import Figlet
from wasabi import table

font = Figlet(font="slant")


def truncate(data, length=75):
    return (data[:length] + '...') if len(data) > length else data


def print_table(data, header, widths, bg_colors=None, aligns=("l", "l", "l")):
    formatted = table(
        data,
        header=header,
        widths=widths,
        aligns=aligns,
        bg_colors=bg_colors,
        divider=True,
    )
    click.secho(formatted)



import base64
import hashlib
import secrets
import time
import uuid
from datetime import datetime, timedelta
from typing import Any, Dict, Generic, List, Optional, TypeVar, cast
from urllib.parse import urlencode

import httpx
import jwt
from jwt.algorithms import RSAAlgorithm
from pydantic import AnyUrl, BaseModel

from kt2.errors import AuthorizeError, GetAccessTokenError
from kt2.jwks import JWKS

algorithm = RSAAlgorithm(RSAAlgorithm.SHA384)


def generate_code_challenge(code_verifier: str) -> str:
    h = hashlib.sha256()
    h.update(code_verifier.encode(encoding='ascii'))
    sha256_val = h.digest()
    code_challenge = bytes.decode(base64.urlsafe_b64encode(sha256_val))
    return code_challenge.replace("+", "-").replace("/", "_").replace("=", "")


class AuthorizeParams(BaseModel):
    aud: str
    launch: str
    client_id: str
    redirect_uri: str
    scope: str
    state: str
    code_challenge: str
    code_challenge_method: str = "S256"
    response_type: str = "code"


class TokenPayload(BaseModel):
    iss: str
    sub: str
    exp: datetime
    aud: AnyUrl
    jti: str


class OIDCTokenBodyAssert(BaseModel):
    client_assertion: str
    redirect_uri: str
    code_verifier: str
    code: str
    client_assertion_type: str = (
        'urn:ietf:params:oauth:client-assertion-type:jwt-bearer'
    )
    grant_type: str = 'authorization_code'


class TokenBodyAssert(BaseModel):
    client_assertion: str
    aud: AnyUrl
    client_assertion_type: str = (
        'urn:ietf:params:oauth:client-assertion-type:jwt-bearer'
    )
    grant_type: str = 'client_credentials'
    scope: str = ""


class OIDCToken(BaseModel):
    access_token: str
    id_token: str
    scope: str
    token_type: str
    expires_in: str
    resource: str
    definition: str
    sub: str
    patient: str
    intent: str


class OAuth2Token(Dict[str, Any]):
    def __init__(self, token_dict: Dict[str, Any]):
        if "expires_at" in token_dict:
            token_dict["expires_at"] = int(token_dict["expires_at"])
        elif "expires_in" in token_dict:
            token_dict["expires_at"] = int(time.time()) + int(token_dict["expires_in"])
        super().__init__(token_dict)

    @property
    def bearer_token(self) -> str:
        access_token = self.get("access_token", None)
        if not access_token:
            return ""

        print("token headers", jwt.get_unverified_header(access_token))
        print(
            "unverified token",
            jwt.decode(access_token, options={"verify_signature": False}),
        )
        return f'Bearer {access_token}'

    def is_expired(self):
        if "expires_at" not in self:
            return True
        return time.time() > self["expires_at"]


T = TypeVar("T")


class SmartFHIROAuth2(Generic[T]):
    name: str
    client_id: str
    authorize_endpoint: str
    access_token_endpoint: str
    jwks: JWKS
    introspection_token_endpoint: Optional[str]
    base_scopes: List[str] = []
    request_headers: Dict[str, str]
    enable_logger: bool = False

    def __init__(
        self,
        client_id: str,
        authorize_endpoint: str,
        access_token_endpoint: str,
        fhir_url: str,
        jwks: JWKS,
        introspection_token_endpoint: Optional[str] = None,
        name: str = "smart-on-fhir",
    ):
        self.client_id = client_id
        self.authorize_endpoint = authorize_endpoint
        self.access_token_endpoint = access_token_endpoint
        self.fhir_url = fhir_url
        self.jwks = jwks
        self.introspection_token_endpoint = introspection_token_endpoint
        self.name = name

        self.token: OAuth2Token = None

        self.request_headers = {
            "Accept": "application/json",
        }

    @property
    def token(self) -> OAuth2Token:
        return self._token

    @token.setter
    def token(self, value) -> None:
        self._token = value

    def get_authorization_url(
        self,
        redirect_uri: str,
        iss: str,
        launch: str,
        state: str,
        code_verifier: str,
        scope: Optional[List[str]] = "launch openid fhirUser",
    ) -> str:
        try:
            params = AuthorizeParams(
                aud=iss,
                launch=launch,
                client_id=self.client_id,
                redirect_uri=redirect_uri,
                code_challenge=generate_code_challenge(code_verifier),
                scope=scope,
                state=state,
            )

            return f"{self.authorize_endpoint}?{urlencode(params.dict())}"
        except:
            raise AuthorizeError(f"Cannot building authorize request: {params.dict()}")

    async def get_soph_access_token(
        self, redirect_uri: str, code_verifier: str, code: str
    ) -> OIDCToken:
        """
        Generate OIDC token on Smart FHIR server

        Raises:
            GetAccessTokenError: token request failed

        Returns:
            OIDCToken: token

        """

        async with httpx.AsyncClient() as client:
            payload = TokenPayload(
                iss=self.client_id,
                sub=self.client_id,
                exp=datetime.now() + timedelta(minutes=5),
                aud=self.access_token_endpoint,
                jti=str(uuid.uuid4().hex),
            )

            token = jwt.encode(
                payload.dict(),
                self.jwks.private_key,
                algorithm="RS512",
                headers={"alg": "RS512", "kid": self.jwks.kid, "typ": "JWT"},
            )

            data = OIDCTokenBodyAssert(
                client_assertion=token,
                redirect_uri=redirect_uri,
                code_verifier=code_verifier,
                code=code,
            )

            token_headers = {
                "Content-Type": "application/x-www-form-urlencoded",
                **self.request_headers,
            }

            self.logger.info("== SMART OIDC AcccesToken ==")
            self.logger.info(f"data: {data.dict()}")
            self.logger.info(f"url:  {self.access_token_endpoint}")
            self.logger.info(f"headers: {token_headers}")

            response = await client.post(
                self.access_token_endpoint,
                data=data.dict(),
                headers=token_headers,
            )

            data = cast(OIDCToken, response.json())

            if response.status_code == 400:
                raise GetAccessTokenError(data)

            token = OIDCToken(**data)

            return token

    async def get_access_token(self) -> OAuth2Token:
        """
        Generate token on Smart FHIR server

        Raises:
            GetAccessTokenError: token request failed

        Returns:
            OAuth2Token: token
        """

        async with httpx.AsyncClient() as client:
            payload = TokenPayload(
                iss=self.client_id,
                sub=self.client_id,
                exp=datetime.now() + timedelta(minutes=5),
                aud=self.access_token_endpoint,
                jti=str(uuid.uuid4().hex),
            )

            token = jwt.encode(
                payload.dict(),
                self.jwks.private_key,
                algorithm="RS512",
                headers={"alg": "RS512", "kid": self.jwks.kid, "typ": "JWT"},
            )

            data = TokenBodyAssert(client_assertion=token, aud=self.fhir_url)

            token_headers = {
                "Content-Type": "application/x-www-form-urlencoded",
                **self.request_headers,
            }

            self.logger.info("== SMART FHIR AcccesToken ==")
            self.logger.info(f"data: {data.dict()}")
            self.logger.info(f"url:  {self.access_token_endpoint}")
            self.logger.info(f"headers: {token_headers}")

            response = await client.post(
                self.access_token_endpoint,
                data=data.dict(),
                headers=token_headers,
            )

            data = cast(Dict[str, Any], response.json())

            if response.status_code == 400:
                raise GetAccessTokenError(data)

            # Set token on client and return for further processing
            token = OAuth2Token(data)

            # Store on client
            self.token = token

            return token

    async def introspect_token(self, access_token: str):
        ...


OAuth2 = SmartFHIROAuth2[Dict[str, Any]]



import click

from kt2.commands import KoppeltaalCli, pass_koppeltaal


@click.group()
@click.option("--debug", is_flag=True, help="enable debug logs ( default: False )")
@click.option("--config", default="koppeltaal.toml", help="select config file")
@click.pass_context
def cli(ctx, **kwargs):
    """Koppeltaal command line tool"""
    ctx.obj = KoppeltaalCli(**kwargs)


@cli.command("patients", short_help="get all patient resources from koppeltaal api")
@pass_koppeltaal
def patients(koppeltaal: KoppeltaalCli, **kwargs):
    koppeltaal.patients(**kwargs)


@cli.command(
    "patient", short_help="get single patient resource by id from koppeltaal api"
)
@click.option("--id", help="id of the patient", required=True)
@pass_koppeltaal
def patient(koppeltaal: KoppeltaalCli, **kwargs):
    koppeltaal.patient(**kwargs)


@cli.command(
    "practitioners", short_help="get all practitioner resources from koppeltaal api"
)
@pass_koppeltaal
def practitioners(koppeltaal: KoppeltaalCli, **kwargs):
    koppeltaal.practitioners(**kwargs)


@cli.command(
    "practitioner",
    short_help="get single practitioner resource by id from koppeltaal api",
)
@click.option("--id", help="id of the practitioner", required=True)
@pass_koppeltaal
def practitioner(koppeltaal: KoppeltaalCli, **kwargs):
    koppeltaal.practitioner(**kwargs)


@cli.command(
    "audit_events", short_help="get all auditevent resources from koppeltaal api"
)
@pass_koppeltaal
def audit_events(koppeltaal: KoppeltaalCli, **kwargs):
    koppeltaal.audit_events(**kwargs)


@cli.command("devices", short_help="get all device resources from koppeltaal api")
@pass_koppeltaal
def endpoints(koppeltaal: KoppeltaalCli, **kwargs):
    koppeltaal.devices(**kwargs)


@cli.command("endpoints", short_help="get all endpoint resources from koppeltaal api")
@pass_koppeltaal
def endpoints(koppeltaal: KoppeltaalCli, **kwargs):
    koppeltaal.endpoints(**kwargs)


@cli.command(
    "endpoint",
    short_help="get single endpoint resource by id from koppeltaal api",
)
@click.option("--id", help="id of the endpoint", required=True)
@pass_koppeltaal
def endpoint(koppeltaal: KoppeltaalCli, **kwargs):
    koppeltaal.endpoint(**kwargs)


@cli.command(
    "activitydefinitions",
    short_help="get all activitydefinition resources from koppeltaal api",
)
@pass_koppeltaal
def activitydefinitions(koppeltaal: KoppeltaalCli, **kwargs):
    koppeltaal.activitydefinitions(**kwargs)


@cli.command(
    "activitydefinition",
    short_help="get single activitydefinition resource by id from koppeltaal api",
)
@click.option("--id", help="id of the activitydefinition", required=True)
@pass_koppeltaal
def activitydefinition(koppeltaal: KoppeltaalCli, **kwargs):
    koppeltaal.activitydefinition(**kwargs)


@cli.command(
    "delete_activitydefinition",
    short_help="get single activitydefinition resource by id from koppeltaal api",
)
@click.option("--id", help="id of the activitydefinition", required=True)
@pass_koppeltaal
def delete_activitydefinition(koppeltaal: KoppeltaalCli, **kwargs):
    koppeltaal.delete_activitydefinition(**kwargs)


@cli.command(
    "tasks",
    short_help="get all task resources from koppeltaal api",
)
@pass_koppeltaal
def tasks(koppeltaal: KoppeltaalCli, **kwargs):
    koppeltaal.tasks(**kwargs)


@cli.command(
    "task",
    short_help="get single task resource by id from koppeltaal api",
)
@click.option("--id", help="id of the task", required=True)
@pass_koppeltaal
def task(koppeltaal: KoppeltaalCli, **kwargs):
    koppeltaal.task(**kwargs)


@cli.command(
    "subscriptions", short_help="get all subscription resources from koppeltaal api"
)
@pass_koppeltaal
def subscriptions(koppeltaal: KoppeltaalCli, **kwargs):
    koppeltaal.subscriptions(**kwargs)


@cli.command("info", short_help="show Koppeltaal api info")
@pass_koppeltaal
def info(koppeltaal: KoppeltaalCli):
    koppeltaal.info()


@cli.command("version", short_help="show Koppeltaal cli version")
@pass_koppeltaal
def version(koppeltaal: KoppeltaalCli):
    koppeltaal.show_version()


if __name__ == "__main__":
    cli()



# Moved from: koppeltaal.jwks

import json
from pathlib import Path

from jwt import PyJWK, PyJWKClient
from jwt.algorithms import RSAAlgorithm
from jwt.types import JWKDict

from kt2.settings import JWKSKeyPair


def get_signing_key_from_jwks(self, kid: str, url: str) -> PyJWK:
    """
    Get signing from .well-known endpoint and generate validation key

    Returns:
        PyJWK: key for decoding JWT tokens
    """
    jwks_client = PyJWKClient(url)
    return jwks_client.get_signing_key(kid=kid)


class JWKS:
    def __init__(self, keys: JWKSKeyPair) -> None:
        self.keys = keys
        self.kid = keys.kid
        self.private_key = None
        self.public_key = None
        self.algorithm = RSAAlgorithm(RSAAlgorithm.SHA384)

    def json(self) -> JWKDict:
        """
        Expose public key in jwks format

        Returns:
            JWKDict: JSON Web Key format
        """
        jwk = json.loads(self.algorithm.to_jwk(self.public_key))
        jwk['kid'] = self.kid
        return jwk

    def load_key_pair(self) -> None:
        private_key_path = Path(self.keys.private_key_path)
        public_key_path = Path(self.keys.public_key_path)

        try:
            private_key_path.resolve(strict=True)
            public_key_path.resolve(strict=True)
        except FileNotFoundError:
            raise FileNotFoundError("Keypair incomplete or not enabled on server")
        else:
            with open(private_key_path) as keyfile:
                self.private_key = self.algorithm.prepare_key(keyfile.read())

            with open(public_key_path) as keyfile:
                self.public_key = self.algorithm.prepare_key(keyfile.read())



from pathlib import Path
from typing import Any, Union

import toml
from pydantic import BaseModel

config = None


class JWKSKeyPair(BaseModel):
    private_key_path: str
    public_key_path: str
    kid: str


class CommandLineArgs(BaseModel):
    # Base config
    config: str = "koppeltaal.toml"
    debug: bool = False


class ConfigFile(CommandLineArgs):
    client_name: str
    fhir_url: str
    oauth_token_url: str
    oauth_authorize_url: str
    oauth_introspection_token_url: str
    smart_config_url: str
    domain: str
    client_id: str
    jwks_keys: JWKSKeyPair


def load_settings(file_path: str) -> Union[Any, dict[str, Any]]:
    path = Path(file_path)
    if path.is_file():
        with open("koppeltaal.toml", "r") as config_file:
            return toml.loads(config_file.read())


def merge_config(config, custom_args) -> ConfigFile:
    # Merge commandline args with presedent over toml file args, except if there None
    filtered_args = {k: v for k, v in custom_args.items() if v is not None}
    return ConfigFile(**{**config, **filtered_args})



from fhir.resources.R4B.fhirtypes import Id, IdentifierType

from kt2.models import (
    KT2ActivityDefinition,
    KT2AuditEvent,
    KT2CareTeam,
    KT2Device,
    KT2Endpoint,
    KT2Organization,
    KT2Patient,
    KT2Practitioner,
    KT2Subscription,
    KT2Task,
)

KT2AuditEventList = list[KT2AuditEvent]
KT2ActivityDefinitionList = list[KT2ActivityDefinition]
KT2CareTeamList = list[KT2CareTeam]
KT2DeviceList = list[KT2Device]
KT2EndpointList = list[KT2Endpoint]
KT2OrganizationList = list[KT2Organization]
KT2PatientList = list[KT2Patient]
KT2PracticionerList = list[KT2Practitioner]
KT2TaskList = list[KT2Task]
KT2SubscriptionList = list[KT2Subscription]
KT2Id = Id
KT2IdentifierType = IdentifierType



from typing import Any, Awaitable, Dict, Optional

import httpx
from httpx._types import HeaderTypes

from kt2 import __version__
from kt2.jwks import JWKS
from kt2.logging import koppeltaal_logger, noop_koppeltaal_logger
from kt2.models import (
    KT2ActivityDefinition,
    KT2AuditEvent,
    KT2CareTeam,
    KT2Device,
    KT2Endpoint,
    KT2Patient,
    KT2Practitioner,
    KT2Subscription,
    KT2Task,
)
from kt2.oauth import SmartFHIROAuth2
from kt2.settings import JWKSKeyPair
from kt2.types import (
    KT2ActivityDefinitionList,
    KT2AuditEventList,
    KT2CareTeamList,
    KT2DeviceList,
    KT2EndpointList,
    KT2Id,
    KT2PatientList,
    KT2PracticionerList,
    KT2SubscriptionList,
    KT2TaskList,
)
from kt2.utils import Methods, make_request, url


class KoppeltaalSDK(SmartFHIROAuth2[Dict[str, Any]]):
    """
    Example scopes:
    ActivityDefinition.cru
    Task.cru
    Patient.r
    Practitioner.r
    Endpoint.r
    """

    def __init__(
        self,
        client_name: str,
        fhir_url: str,
        oauth_token_url: str,
        oauth_authorize_url: str,
        domain: str,
        client_id: str,
        smart_config_url: str,
        jwks_keys: JWKSKeyPair,
        oauth_introspection_token_url: Optional[str] = None,
        connector_name: str = "kt2-python-sdk",
        enable_logger=False,
    ):
        self.connector_name = connector_name
        self.name: str = client_name
        self.fhir_url: str = fhir_url
        self.smart_config_url: str = smart_config_url
        self.domain: str = domain
        self.logger = koppeltaal_logger if enable_logger else noop_koppeltaal_logger

        # Do not define JWKS on client: instead pass on the object to SmartFHIROAuth2

        jwks = JWKS(jwks_keys)
        jwks.load_key_pair()

        super().__init__(
            client_id=client_id,
            authorize_endpoint=oauth_authorize_url,
            access_token_endpoint=oauth_token_url,
            fhir_url=fhir_url,
            jwks=jwks,
            introspection_token_endpoint=oauth_introspection_token_url,
        )

        self.request_headers = {
            "Accept": "application/fhir+json",
            "User-Agent": f"{self.connector_name}/{__version__}",
        }

    def default_headers(self, headers: HeaderTypes = {}) -> HeaderTypes:
        """Set default headers"""
        return {
            **self.request_headers,
            **{"Authorization": self.token.bearer_token},
            **headers,
        }

    def post_headers(self, id: KT2Id) -> HeaderTypes:
        print("id =", id)
        return self.default_headers(
            headers={
                # "If-None-Exist": id,
                "Prefer": "return=representation",
                "Content-Type": "application/fhir+json",
            }
        )

    def put_headers(self, version_id: KT2Id) -> HeaderTypes:
        return self.default_headers(
            headers={
                "If-Match": f'W/"{version_id}"',
                "Prefer": "return=representation",
                "Content-Type": "application/fhir+json",
            }
        )

    def subscription_headers(self) -> HeaderTypes:
        return self.default_headers(
            headers={
                "Prefer": "return=representation",
                "Content-Type": "application/fhir+json",
            }
        )

    async def get_info(self, *args, **kwargs) -> None:
        """
        Get info from Koppeltaal Api
        """

        # FIXME: add SmartConfig Model so we can view data better
        async with httpx.AsyncClient(headers=self.request_headers) as client:
            return await make_request(
                client,
                method=Methods.GET,
                url=self.smart_config_url,
                logger=self.logger,
            )

    @url("/Patient", response_model=KT2Patient, many=True)
    async def get_patients(self, *args, **kwargs) -> Awaitable[KT2PatientList]:
        """
        Get all Patient resources from Koppeltaal Api
        """

        async with httpx.AsyncClient(headers=self.default_headers()) as client:
            return await make_request(
                client,
                method=Methods.GET,
                url=kwargs["url"],
                logger=self.logger,
            )

    @url("/Patient", response_model=KT2Patient)
    async def get_patient(self, id: str, *args, **kwargs) -> Awaitable[KT2Patient]:
        """
        Get single Patient resource by id from Koppeltaal Api
        """

        async with httpx.AsyncClient(headers=self.default_headers()) as client:
            return await make_request(
                client,
                method=Methods.GET,
                url=f'{kwargs["url"]}/{id}',
                logger=self.logger,
            )

    @url("/Practitioner", response_model=KT2Practitioner, many=True)
    async def get_practitioners(
        self, *args, **kwargs
    ) -> Awaitable[KT2PracticionerList]:
        """
        Get all Practitioner resources from Koppeltaal Api
        """

        async with httpx.AsyncClient(headers=self.default_headers()) as client:
            return await make_request(
                client,
                method=Methods.GET,
                url=kwargs["url"],
                logger=self.logger,
            )

    @url("/Practitioner", response_model=KT2Practitioner)
    async def get_practitioner(
        self, id: str, *args, **kwargs
    ) -> Awaitable[KT2Practitioner]:
        """
        Get single Practitioner resource by id from Koppeltaal Api
        """

        async with httpx.AsyncClient(headers=self.default_headers()) as client:
            return await make_request(
                client,
                method=Methods.GET,
                url=f'{kwargs["url"]}/{id}',
                logger=self.logger,
            )

    @url(
        "/AuditEvent",
        response_model=KT2AuditEvent,
        many=True,
    )
    async def get_audit_events(self, *args, **kwargs) -> Awaitable[KT2AuditEventList]:
        """
        Get all AuditEvent resources from Koppeltaal Api

        Example: AuditEvent?page=3&queryId=2185a414-a0b5-11ee-a9c8-4a9017cfa52e
        """

        async with httpx.AsyncClient(headers=self.default_headers()) as client:
            return await make_request(
                client,
                method=Methods.GET,
                url=kwargs["url"],
                logger=self.logger,
            )

    @url("/AuditEvent", response_model=KT2AuditEvent)
    async def create_audit_event(
        self, audit_event: KT2AuditEvent, *args, **kwargs
    ) -> Awaitable[KT2AuditEvent]:
        """
        Create AuditEvent resource in Koppeltaal Api
        """

        async with httpx.AsyncClient(
            headers=self.post_headers(audit_event.id)
        ) as client:
            return await make_request(
                client,
                method=Methods.POST,
                url=kwargs["url"],
                data=audit_event.json(exclude_none=True, exclude_comments=True),
                logger=self.logger,
            )

    @url("/Device", response_model=KT2Device, many=True)
    async def get_devices(self, *args, **kwargs) -> Awaitable[KT2DeviceList]:
        """
        Get all Device resources from Koppeltaal Api
        """

        async with httpx.AsyncClient(headers=self.default_headers()) as client:
            return await make_request(
                client,
                method=Methods.GET,
                url=kwargs["url"],
                logger=self.logger,
            )

    @url("/Endpoint", response_model=KT2Endpoint, many=True)
    async def get_endpoints(self, *args, **kwargs) -> Awaitable[KT2EndpointList]:
        """
        Get all Endpoint resources from Koppeltaal Api
        """

        async with httpx.AsyncClient(headers=self.default_headers()) as client:
            return await make_request(
                client,
                method=Methods.GET,
                url=kwargs["url"],
                logger=self.logger,
            )

    @url("/Endpoint", response_model=KT2Endpoint)
    async def get_endpoint(self, id: str, *args, **kwargs) -> Awaitable[KT2Endpoint]:
        """
        Get single Endpoint resource by id from Koppeltaal Api
        """

        async with httpx.AsyncClient(headers=self.default_headers()) as client:
            return await make_request(
                client,
                method=Methods.GET,
                url=f'{kwargs["url"]}/{id}',
                logger=self.logger,
            )

    @url("/Endpoint", response_model=KT2Endpoint)
    async def create_endpoint(
        self, endpoint: KT2Endpoint, *args, **kwargs
    ) -> Awaitable[KT2Endpoint]:
        """
        Create Endpoint resource in Koppeltaal Api
        """

        async with httpx.AsyncClient(headers=self.post_headers(endpoint.id)) as client:
            return await make_request(
                client,
                method=Methods.POST,
                url=kwargs["url"],
                data=endpoint.json(exclude_none=True, exclude_comments=True),
                logger=self.logger,
            )

    @url("/CareTeam", response_model=KT2CareTeam, many=True)
    async def get_careteams(self, *args, **kwargs) -> Awaitable[KT2CareTeamList]:
        """
        Get all CareTeam resources from Koppeltaal Api
        """

        async with httpx.AsyncClient(headers=self.default_headers()) as client:
            return await make_request(
                client,
                method=Methods.GET,
                url=kwargs["url"],
                logger=self.logger,
            )

    @url("/CareTeam", response_model=KT2CareTeam)
    async def get_careteam(self, id: str, *args, **kwargs) -> Awaitable[KT2CareTeam]:
        """
        Get single CareTeam resource by id from Koppeltaal Api
        """

        async with httpx.AsyncClient(headers=self.default_headers()) as client:
            return await make_request(
                client,
                method=Methods.GET,
                url=f'{kwargs["url"]}/{id}',
                logger=self.logger,
            )

    @url("/ActivityDefinition", response_model=KT2ActivityDefinition, many=True)
    async def get_activitydefinitions(
        self, *args, **kwargs
    ) -> Awaitable[KT2ActivityDefinitionList]:
        """
        Get all ActivityDefinition resources from Koppeltaal Api
        """

        async with httpx.AsyncClient(headers=self.default_headers()) as client:
            return await make_request(
                client,
                method=Methods.GET,
                url=kwargs["url"],
                logger=self.logger,
            )

    @url("/ActivityDefinition", response_model=KT2ActivityDefinition)
    async def get_activitydefinition(
        self, id: str, *args, **kwargs
    ) -> Awaitable[KT2ActivityDefinition]:
        """
        Get single ActivityDefinition resource by id from Koppeltaal Api
        """

        async with httpx.AsyncClient(headers=self.default_headers()) as client:
            return await make_request(
                client,
                method=Methods.GET,
                url=f'{kwargs["url"]}/{id}',
                logger=self.logger,
            )

    @url("/ActivityDefinition", response_model=KT2ActivityDefinition)
    async def create_activitydefinition(
        self, activitydefinition: KT2ActivityDefinition, *args, **kwargs
    ) -> Awaitable[KT2ActivityDefinition]:
        """
        Create ActivityDefinition resource in Koppeltaal Api
        """

        async with httpx.AsyncClient(
            headers=self.post_headers(activitydefinition.id)
        ) as client:
            return await make_request(
                client,
                method=Methods.POST,
                url=kwargs["url"],
                data=activitydefinition.json(exclude_none=True, exclude_comments=True),
                logger=self.logger,
            )

    @url("/ActivityDefinition", response_model=KT2ActivityDefinition)
    async def update_activitydefinition(
        self, activitydefinition: KT2ActivityDefinition, *args, **kwargs
    ) -> Awaitable[KT2ActivityDefinition]:
        """
        Update ActivityDefinition resource in Koppeltaal Api
        """

        async with httpx.AsyncClient(
            headers=self.put_headers(activitydefinition.meta.versionId)
        ) as client:
            return await make_request(
                client,
                method=Methods.PUT,
                url=f'{kwargs["url"]}/{activitydefinition.id}',
                data=activitydefinition.json(exclude_none=True, exclude_comments=True),
                logger=self.logger,
            )

    @url("/ActivityDefinition", response_model=KT2ActivityDefinition)
    async def delete_activitydefinition(
        self, id: str, *args, **kwargs
    ) -> Awaitable[KT2ActivityDefinition]:
        """
        Get delete single ActivityDefinition resource by id from Koppeltaal Api
        """

        async with httpx.AsyncClient(headers=self.default_headers()) as client:
            return await make_request(
                client,
                method=Methods.DELETE,
                url=f'{kwargs["url"]}/{id}',
                logger=self.logger,
            )

    @url("/Task", response_model=KT2Task, many=True)
    async def get_tasks(self, *args, **kwargs) -> Awaitable[KT2TaskList]:
        """
        Get all Task resources from Koppeltaal Api
        """

        async with httpx.AsyncClient(headers=self.default_headers()) as client:
            return await make_request(
                client,
                method=Methods.GET,
                url=kwargs["url"],
                logger=self.logger,
            )

    @url("/Task", response_model=KT2Task)
    async def get_task(self, id: str, *args, **kwargs) -> Awaitable[KT2Task]:
        """
        Get single Task resource by id from Koppeltaal Api
        """

        async with httpx.AsyncClient(headers=self.default_headers()) as client:
            return await make_request(
                client,
                method=Methods.GET,
                url=f'{kwargs["url"]}/{id}',
                logger=self.logger,
            )

    @url("/Task", response_model=KT2Task)
    async def create_task(self, task: KT2Task, *args, **kwargs) -> Awaitable[KT2Task]:
        """
        Create Task resource in Koppeltaal Api
        """

        async with httpx.AsyncClient(headers=self.post_headers(task.id)) as client:
            return await make_request(
                client,
                method=Methods.POST,
                url=kwargs["url"],
                data=task.json(exclude_none=True, exclude_comments=True),
                logger=self.logger,
            )

    @url("/Task", response_model=KT2Task)
    async def update_task(self, task: KT2Task, *args, **kwargs) -> Awaitable[KT2Task]:
        """
        Update Task resource in Koppeltaal Api
        """

        async with httpx.AsyncClient(
            headers=self.put_headers(task.meta.versionId)
        ) as client:
            return await make_request(
                client,
                method=Methods.PUT,
                url=f'{kwargs["url"]}/{task.id}',
                data=task.json(exclude_none=True, exclude_comments=True),
                logger=self.logger,
            )

    @url("/Subscription", response_model=KT2Subscription, many=True)
    async def get_subscriptions(
        self, *args, **kwargs
    ) -> Awaitable[KT2SubscriptionList]:
        """
        Get all Subscription resources from Koppeltaal Api
        """

        async with httpx.AsyncClient(headers=self.default_headers()) as client:
            return await make_request(
                client,
                method=Methods.GET,
                url=kwargs["url"],
                logger=self.logger,
            )

    @url("/Subscription", response_model=KT2Subscription)
    async def subscribe(
        self, subscription: KT2Subscription, *args, **kwargs
    ) -> Awaitable[KT2Subscription]:
        """
        Subscription resource in Koppeltaal Api
        """

        async with httpx.AsyncClient(headers=self.subscription_headers()) as client:
            return await make_request(
                client,
                method=Methods.POST,
                url=kwargs["url"],
                data=subscription.json(exclude_none=True, exclude_comments=True),
                logger=self.logger,
            )



import typing

from fhir.resources.R4B import fhirtypes
from fhir.resources.R4B.activitydefinition import ActivityDefinition
from fhir.resources.R4B.auditevent import AuditEventAgent  # noqa
from fhir.resources.R4B.auditevent import AuditEventEntity  # noqa
from fhir.resources.R4B.auditevent import AuditEventSource  # noqa
from fhir.resources.R4B.auditevent import AuditEvent
from fhir.resources.R4B.careteam import CareTeam, CareTeamParticipant
from fhir.resources.R4B.codeableconcept import CodeableConcept  # noqa
from fhir.resources.R4B.coding import Coding  # noqa
from fhir.resources.R4B.device import Device  # noqa
from fhir.resources.R4B.endpoint import Endpoint
from fhir.resources.R4B.extension import Extension  # noqa
from fhir.resources.R4B.identifier import Identifier  # noqa
from fhir.resources.R4B.meta import Meta
from fhir.resources.R4B.organization import Organization
from fhir.resources.R4B.patient import Patient
from fhir.resources.R4B.practitioner import Practitioner
from fhir.resources.R4B.reference import Reference  # noqa
from fhir.resources.R4B.subscription import Subscription, SubscriptionChannel  # noqa
from fhir.resources.R4B.task import Task
from pydantic import Field

if typing.TYPE_CHECKING:
    from pydantic.types import CallableGenerator


class PersonMixin:
    @property
    def primary_email(self) -> str:
        return next(
            (contact.value for contact in self.telecom if contact.system == 'email'), ''
        )

    @property
    def first_name(self) -> str:
        return " ".join(
            next((humanname.given for humanname in self.name if humanname.given), '')
        )

    @property
    def middle_name(self) -> str:
        return " ".join(
            next((humanname.prefix for humanname in self.name if humanname.prefix), '')
        )

    @property
    def last_name(self) -> str:
        return next(
            (humanname.family for humanname in self.name if humanname.family), ''
        )

    @property
    def display_name(self) -> str:
        names = [self.first_name, self.middle_name, self.last_name]
        return ' '.join([name for name in names if name])


class KT2ActivityDefinition(ActivityDefinition):
    """
    Koppeltaal structure of ActivityDefinition
    """

    meta: fhirtypes.MetaType = Field(
        Meta(
            profile=[
                "http://koppeltaal.nl/fhir/StructureDefinition/KT2ActivityDefinition"
            ]
        ),
        alias="meta",
        title="Metadata about the resource",
        description=(
            "The metadata about the resource. This is content that is maintained by"
            " the infrastructure. Changes to the content might not always be "
            "associated with version changes to the resource."
        ),
        # if property is element of this resource.
        element_property=True,
    )


class KT2Task(Task):
    """
    Koppeltaal structure of Task

    Changes:

    - Task.partOf `KT2Task`: Reference(KT2Task) ..
    - Task.for `KT2Patient`: Reference(KT2Patient) 1..
    - Task.requester `KT2Practitioner`: Reference(KT2Practitioner) ..
    - Task.owner `KT2CareTeam | KT2Patient | KT2Practitioner`: Reference(KT2CareTeam | KT2Patient | KT2Practitioner) 1..
    """

    meta: fhirtypes.MetaType = Field(
        Meta(profile=["http://koppeltaal.nl/fhir/StructureDefinition/KT2Task"]),
        alias="meta",
        title="Metadata about the resource",
        description=(
            "The metadata about the resource. This is content that is maintained by"
            " the infrastructure. Changes to the content might not always be "
            "associated with version changes to the resource."
        ),
        # if property is element of this resource.
        element_property=True,
    )

    partOf: typing.List[fhirtypes.ReferenceType] = Field(
        None,
        alias="partOf",
        title="Composite task",
        description="Task that this particular task is part of.",
        # if property is element of this resource.
        element_property=True,
        # note: Listed Resource Type(s) should be allowed as Reference.
        enum_reference_types=["KT2Task"],
    )

    for_fhir: fhirtypes.ReferenceType = Field(
        None,
        alias="for",
        title="Beneficiary of the Task",
        description=(
            "The entity who benefits from the performance of the service specified "
            "in the task (e.g., the patient)."
        ),
        # if property is element of this resource.
        element_property=True,
        # note: Listed Resource Type(s) should be allowed as Reference.
        enum_reference_types=["KT2Patient"],
    )

    requester: fhirtypes.ReferenceType = Field(
        None,
        alias="requester",
        title="Who is asking for task to be done",
        description="The creator of the task.",
        # if property is element of this resource.
        element_property=True,
        # note: Listed Resource Type(s) should be allowed as Reference.
        enum_reference_types=["KT2Practitioner"],
    )

    owner: fhirtypes.ReferenceType = Field(
        None,
        alias="owner",
        title="Responsible individual",
        description="Party responsible for managing task execution.",
        # if property is element of this resource.
        element_property=True,
        # note: Listed Resource Type(s) should be allowed as Reference.
        enum_reference_types=["KT2CareTeam", "KT2Patient", "KT2Practitioner"],
    )


class KT2Endpoint(Endpoint):
    """
    Koppeltaal structure of Endpoint

    Changes:
    - Endpoint.managingOrganization `KT2Organization`: Reference(KT2Organization) ..
    """

    meta: fhirtypes.MetaType = Field(
        Meta(profile=["http://koppeltaal.nl/fhir/StructureDefinition/KT2Endpoint"]),
        alias="meta",
        title="Metadata about the resource",
        description=(
            "The metadata about the resource. This is content that is maintained by"
            " the infrastructure. Changes to the content might not always be "
            "associated with version changes to the resource."
        ),
        # if property is element of this resource.
        element_property=True,
    )

    managingOrganization: fhirtypes.ReferenceType = Field(
        None,
        alias="managingOrganization",
        title=(
            "Organization that manages this endpoint (might not be the organization"
            " that exposes the endpoint)"
        ),
        description=(
            "The organization that manages this endpoint (even if technically "
            "another organization is hosting this in the cloud, it is the "
            "organization associated with the data)."
        ),
        # if property is element of this resource.
        element_property=True,
        # note: Listed Resource Type(s) should be allowed as Reference.
        enum_reference_types=["KT2Organization"],
    )


class KT2CareTeamParticipant(CareTeamParticipant):
    """Koppeltaal structure for CareTeamParticipant

    Changes:
    - CareTeamParticipant.member `KT2Practitioner`: Reference(KT2Practitioner) ..
    """

    meta: fhirtypes.MetaType = Field(
        Meta(
            profile=[
                "http://koppeltaal.nl/fhir/StructureDefinition/KT2CareTeamParticipant"
            ]
        ),
        alias="meta",
        title="Metadata about the resource",
        description=(
            "The metadata about the resource. This is content that is maintained by"
            " the infrastructure. Changes to the content might not always be "
            "associated with version changes to the resource."
        ),
        # if property is element of this resource.
        element_property=True,
    )

    member: fhirtypes.ReferenceType = Field(
        None,
        alias="member",
        title="Who is involved",
        description=(
            "The specific person or organization who is participating/expected to "
            "participate in the care team."
        ),
        # if property is element of this resource.
        element_property=True,
        # note: Listed Resource Type(s) should be allowed as Reference.
        enum_reference_types=["KT2Practitioner"],
    )


class KT2CareTeamParticipantType(fhirtypes.AbstractType):
    """
    Custom abstract type for CareTeamParticipantType

    Args:
        fhirtypes `AbstractType`: KT2CareTeamParticipant AbstractType mapping
    """

    __resource_type__ = "KT2CareTeamParticipant"

    @classmethod
    def __get_validators__(cls) -> "CallableGenerator":
        from fhir.resources import fhirtypesvalidators

        yield getattr(fhirtypesvalidators, "CareTeamParticipant".lower() + "_validator")


class KT2CareTeam(CareTeam):
    """
    Koppeltaal structure of CareTeam

    Changes:
    - CareTeam.subject `KT2Patient`: Reference(KT2Patient) ..
    - CareTeam.participant.member `KT2Practitioner`: Reference(KT2Practitioner) ..
    - CareTeam.managingOrganization `KT2Organization`: Reference(KT2Organization) ..1
    """

    meta: fhirtypes.MetaType = Field(
        Meta(profile=["http://koppeltaal.nl/fhir/StructureDefinition/KT2CareTeam"]),
        alias="meta",
        title="Metadata about the resource",
        description=(
            "The metadata about the resource. This is content that is maintained by"
            " the infrastructure. Changes to the content might not always be "
            "associated with version changes to the resource."
        ),
        # if property is element of this resource.
        element_property=True,
    )

    subject: fhirtypes.ReferenceType = Field(
        None,
        alias="subject",
        title="Who care team is for",
        description=(
            "Identifies the patient or group whose intended care is handled by the "
            "team."
        ),
        # if property is element of this resource.
        element_property=True,
        # note: Listed Resource Type(s) should be allowed as Reference.
        enum_reference_types=["KT2Patient"],
    )

    participant: typing.List[KT2CareTeamParticipantType] = Field(
        None,
        alias="participant",
        title="Members of the team",
        description=(
            "Identifies all people and organizations who are expected to be "
            "involved in the care team."
        ),
        # if property is element of this resource.
        element_property=True,
    )

    managingOrganization: typing.List[fhirtypes.ReferenceType] = Field(
        None,
        alias="managingOrganization",
        title="Organization responsible for the care team",
        description="The organization responsible for the care team.",
        # if property is element of this resource.
        element_property=True,
        # note: Listed Resource Type(s) should be allowed as Reference.
        enum_reference_types=["KT2Organization"],
    )


class KT2Patient(PersonMixin, Patient):
    """
    Koppeltaal structure of Patient

    Changes:
    - Patient.managingOrganization `KT2Organization`: Reference(KT2Organization) ..
    """

    meta: fhirtypes.MetaType = Field(
        Meta(profile=["http://koppeltaal.nl/fhir/StructureDefinition/KT2Patient"]),
        alias="meta",
        title="Metadata about the resource",
        description=(
            "The metadata about the resource. This is content that is maintained by"
            " the infrastructure. Changes to the content might not always be "
            "associated with version changes to the resource."
        ),
        # if property is element of this resource.
        element_property=True,
    )

    managingOrganization: fhirtypes.ReferenceType = Field(
        None,
        alias="managingOrganization",
        title="Organization that is the custodian of the patient record",
        description=None,
        # if property is element of this resource.
        element_property=True,
        # note: Listed Resource Type(s) should be allowed as Reference.
        enum_reference_types=["KT2Organization"],
    )


class KT2Practitioner(PersonMixin, Practitioner):
    """
    Koppeltaal structure of Practitioner

    Changes:
    - No profile changes
    """

    meta: fhirtypes.MetaType = Field(
        Meta(profile=["http://koppeltaal.nl/fhir/StructureDefinition/KT2Practitioner"]),
        alias="meta",
        title="Metadata about the resource",
        description=(
            "The metadata about the resource. This is content that is maintained by"
            " the infrastructure. Changes to the content might not always be "
            "associated with version changes to the resource."
        ),
        # if property is element of this resource.
        element_property=True,
    )


class KT2Organization(Organization):
    """
    Koppeltaal structure of Organization

    Changes:
    - Organization.partOf `KT2Organization`: Reference(KT2Organization) ..
    - Organization.endpoint `KT2Endpoint`: Reference(KT2Endpoint) ..0
    """

    meta: fhirtypes.MetaType = Field(
        Meta(profile=["http://koppeltaal.nl/fhir/StructureDefinition/KT2Organization"]),
        alias="meta",
        title="Metadata about the resource",
        description=(
            "The metadata about the resource. This is content that is maintained by"
            " the infrastructure. Changes to the content might not always be "
            "associated with version changes to the resource."
        ),
        # if property is element of this resource.
        element_property=True,
    )

    partOf: fhirtypes.ReferenceType = Field(
        None,
        alias="partOf",
        title="The organization of which this organization forms a part",
        description=None,
        # if property is element of this resource.
        element_property=True,
        # note: Listed Resource Type(s) should be allowed as Reference.
        enum_reference_types=["KT2Organization"],
    )

    endpoint: typing.List[fhirtypes.ReferenceType] = Field(
        None,
        alias="endpoint",
        title=(
            "Technical endpoints providing access to services operated for the "
            "organization"
        ),
        description=None,
        # if property is element of this resource.
        element_property=True,
        # note: Listed Resource Type(s) should be allowed as Reference.
        enum_reference_types=["KT2Endpoint"],
    )


class KT2Subscription(Subscription):
    meta: fhirtypes.MetaType = Field(
        Meta(profile=["http://koppeltaal.nl/fhir/StructureDefinition/KT2Subscription"]),
        alias="meta",
        title="Metadata about the resource",
        description=(
            "The metadata about the resource. This is content that is maintained by"
            " the infrastructure. Changes to the content might not always be "
            "associated with version changes to the resource."
        ),
        # if property is element of this resource.
        element_property=True,
    )


class KT2SubscriptionChannel(SubscriptionChannel):
    pass


class KT2AuditEvent(AuditEvent):
    meta: fhirtypes.MetaType = Field(
        Meta(profile=["http://koppeltaal.nl/fhir/StructureDefinition/KT2AuditEvent"]),
        alias="meta",
        title="Metadata about the resource",
        description=(
            "The metadata about the resource. This is content that is maintained by"
            " the infrastructure. Changes to the content might not always be "
            "associated with version changes to the resource."
        ),
        # if property is element of this resource.
        element_property=True,
    )


class KT2Device(Device):
    pass
